#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
BW2QSD: Bridging Brightway2 and QSD packages for LCA

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

Part of this module is based on the BioSTEAM-LCA package:
https://github.com/scyjth/biosteam_lca

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/BW2QSD/blob/main/LICENSE.txt
for license details.
'''

import os, appdirs, subprocess, shutil, requests, functools
import brightway2 as bw2
import eidl
from zipfile import ZipFile
from bw2io import importers, strategies

'''
TODO:
    JSON conversion, then add USLCI, etc.
'''


__all__ = ('DataDownloader',)


def _check_db(name, force_skipping=False):
    dbs = []
    for k in bw2.databases.keys():
        if name.lower() in k:
            dbs.append(k)
    
    if dbs:
        print(f'\nThe following {name} database(s) {dbs} exist(s).')

        if force_skipping:
            return False
        else:
            print(f'\nWould you like to download a different version of {name}?')
            if input('[y]/[n]: ') in ('n', 'N', 'no', 'No', 'NO'):
                return False
    
    return True


def _make_dir(path, end_dir=''):
    if not path:
        path = fp = appdirs.user_data_dir(appname='BW2QSD', appauthor='bw2qsd')
        if not os.path.isdir(fp):
            os.makedirs(fp)
            print (f'\nDirectory {fp} created for storing database.')

    full_path = os.path.join(path, end_dir) if end_dir else path
    if not os.path.isdir(full_path):
        os.makedirs(full_path)
        print (f'\nDirectory {full_path} created for storing database.')

    return full_path



class DataDownloader:
    '''
    To download databases from external sources.
    
    Currently support:
        - `ecoinvent <https://www.ecoinvent.org/>`_, version 3+, (license and login credentials required)
        - `FORWAST <https://lca-net.com/projects/show/forwast/>`_
        - `USLCI <>`_ (NOT YET READY)
    
    '''
        
    def download_ecoinvent(self, path='', remove_download=False,
                           remove_cache_data=False):
        '''
        Download ecoinvent database using the ``eidl`` package.
        You will be prompted to enter ecoinvent license and login credentials.
        
        The original package ``eidl`` has compatibility issues with some operating systems,
        e.g., :func:`eidl.get_ecoinvent` may not work, this function fixes the bugs.
        
        .. note::
            
            The entire process may take 10-20 min.
        
        
        Parameters
        ----------
        path : str
            Path for for storing the extracted raw database file.
            Will use the user data storage directory (based on :func:`appdirs.user_data_dir`)
            if not provided.
            
            .. note::
            
                Downloaded files are stored under the "EcoInventDownLoader" directory
                in the user data storage directory.


        remove_download : bool
            Whether to remove the downloaded zipfile after extracting.
        remove_cache_data : bool
            Whether to remove the raw database files (i.e., unzipped zipfile)
            after importing.
        
        Tip
        ---
        The following system models are supported by ecoinvent [1]_:
            
            [1] cutoff: Allocation, cut-off by classification. \
            For users new to LCI databases, ecoinvent recommends using the \
            cut-off system model when starting to work with ecoinvent version 3.
            
            [2] apos: Allocation at the point of substitution.
            
            [3] consequential: Substitution, consequential,


        .. note::
            
            When selecting system models, you may run into both "cut-off" and "cutoff"
            models, choose "cutoff", "cut-off" is a bug in ecoinvent.

        
        References
        ----------
        .. [1] `System Models in ecoinvent 3 <https://www.ecoinvent.org/database/system-models-in-ecoinvent-3/system-models-in-ecoinvent-3.html>`_
        
        
        '''
        if not _check_db('ecoinvent'):
            return

        path = _make_dir(path, 'ecoinvent')
        downloader = eidl.EcoinventDownloader()
        downloader.run()

        print('\nUnzipping data...')
        
        db_append = downloader.file_name.replace('.7z', '')
        
        extracted_path = os.path.join(path, db_append)
        try: # remove previous extracted cache
            os.rmdir(extracted_path)
        except FileNotFoundError:
            pass
        
        extract_cmd = ['7za', 'x', downloader.out_path, f'-o{extracted_path}']
        # # do not use downloader.extract, it may not work on Mac app 
        # downloader.extract(target_dir=path)
        
        self.extraction_process = subprocess.Popen(extract_cmd)
        self.extraction_process.wait()

        db_name = 'ecoinvent_' + db_append
        datasets_path = os.path.join(extracted_path, 'datasets') 
                
        ecospold_import = importers.SingleOutputEcospold2Importer(datasets_path, db_name)
        ecospold_import.apply_strategies()
        
        print('\nInspecting data...')
        self.inspect(ecospold_import, db_name)
        
        if remove_download:
            os.remove(downloader.out_path)

        if remove_cache_data:
            shutil.rmtree(extracted_path)
            try:
                os.rmdir(path)
            except OSError:
                pass
                
        print(f'\nSuccessfully imported ecoinvent database as "ecoinvent_{db_append}".')


    def download_forwast(self, path='',
                         url='http://lca-net.com/wp-content/uploads/forwast.bw2package.zip',
                         remove_download=False, remove_cache_data=False):
        '''
        Download the FORWAST database.
        
        Parameters
        ----------
        path : str
            Path for for storing the downloaded and extracted raw database file.
            Will use the user data storage directory (based on :func:`appdirs.user_data_dir`)
            if not provided.
        url : str
            FORWAST database downloading url.
            You may need to update the url according to the FORWAST website
            if the default one is not working.       
        remove_download : bool
            Whether to remove the downloaded zipfile after extracting.
        remove_cache_data : bool
            Whether to remove the raw database files (i.e., unzipped zipfile)
            after importing.
        
        See Also
        --------
        The `FORWARST project <https://lca-net.com/projects/show/forwast/>`_.
        
        '''
        if not _check_db('FORWAST', True):
            return
        
        path = _make_dir(path, 'forwast')
        fp = os.path.join(path, 'forwast.package.zip')
        fp_extracted = os.path.join(path, 'forwast.bw2package')
        if os.path.exists(fp_extracted):
            print('\nUsing previously extracted FORWAST package in directory ' \
                  f'"{path}".')
        
        else:            
            if os.path.exists(fp):
                print('\nUsing previously downloaded FORWAST package in directory ' \
                      f'"{path}".')
    
            else:
                print('\nDownloading data...\n')
                r = requests.get(url, stream=True)
                if r.status_code != 200:
                    raise (f'URL "{url}" returns status code "{r.status_code}".')
               
                # From BioSTEAM-LCA:
                # use the following code instead of ``r.raw.read`` to save what is being streamed to a file.
                # with open(filename, 'wb') as fd:
                with open(fp, 'wb') as fd:
                    for chunk in r.iter_content(chunk_size=128): # chunk = 128 * 1024
                        fd.write(chunk)
            
            print('\nExtracting data...')
            ZipFile(fp).extractall(path)         
        bw2.BW2Package.import_file(os.path.join(path, 'forwast.bw2package'))
        
        if remove_download:
            os.remove(fp)

        if remove_cache_data:
            shutil.rmtree(path)
    
        print('\nSuccessfully imported FORWAST database as "forwast".')



    def download_USLCI(self, db_path):
        '''
        NOT READY YET
        
        
        Import the pre-downloaded and pre-converted (from JSON to ecoSpold1 or ecoSpold2)
        U.S. Life Cycle Inventory (USLCI) database from
        `Federal LCA Commons repository <https://www.lcacommons.gov/lca-collaboration/>`.
        
        .. note::
            
            The files in the `Federal LCA Commons repository <https://www.lcacommons.gov/lca-collaboration/>`
            are in JSON file, and need to be converted to ecoSpold1 (`outdated <https://www.ecoinvent.org/data-provider/data-provider-toolkit/ecospold2/changes-from-ecospold1-to-ecospold2/changes-from-ecospold1-to-ecospold2.html>`_)
            or ecoSpold2 before using.
            
        Parameters
        ----------
        db_path : str
            Directory to the sown
        
        
        '''
        db_name = 'us_lci'
        
        
        from warnings import warn
        warn('This function is not ready yet.')
        return
    
        #!!! Look into brightway2 for direct importing from JSON
    
        lci_import = importers.SingleOutputEcospold2Importer(
            os.path.join(self.dirpath, 'US_LCI'), db_name)
        lci_import.apply_strategies()
        
        lci_import.migrate('unusual-units')
        lci_import.migrate('default-units')
        
        # Link the biosphere flows by their names, units, and categories
        link_iter = functools.partial(strategies.link_iterable_by_fields, 
                                      other= bw2.Database(bw2.config.biosphere),
                                      kind='biosphere')
        lci_import.apply_strategy(link_iter)
        
        print('\nInspecting data...\n')        
        self.inspect(lci_import, db_name)

        print('\nSuccessfully imported U.S. Life Cycle Inventory database as "us_lci".')


    @staticmethod
    def inspect(sp, db_name):
        '''
        Check for unlinked exchanges in a given database. 
        If found, it will be the user's decision to if or not continue writing to 
        SQLite3 backend.

        Parameters
        ----------
        sp : obj
            The initialized database to be inspected.
        db_name : obj
            Name of the database being inspected.
        
        Returns
        -------
        datasets: int 
            The total number of lci datasets extracted.        
        exchanges : int
            The total number of exchanges.
        unlinked: int
             The total number of unlinked exchanges.
        '''
        
        datasets, exchanges, unlinked = sp.statistics(print_stats=False)
        
        if not unlinked:
            sp.write_database()
        else:
            print(f'\nThere are {unlinked} unlinked exchanges, would you like to show all unlinked exchanges?')
            if input('[y]/[n]: ') in ('y', 'yes', 'Y', 'Yes', 'YES'):
                 for x in sp.unlinked:
                     print(x)

            print(f'\nContinue to write database {db_name}?')            
            if input('[y]/[n]: ') in ('y', 'yes', 'Y', 'Yes', 'YES'):
                print ('Deleting exchanges with zero amount...')
                for ds in sp.data:
                    ds['exchanges'] = [exc for exc in ds['exchanges'] if (exc['amount'] or exc['uncertainty type'] != 0)]
                
                print ('\nDrop unlinked exchanges?')
                if input('[y]/[n]: ') in ('y', 'yes', 'Y', 'Yes', 'YES'):
                    try:
                        sp.apply_strategies([strategies.generic.drop_unlinked])  #sp.drop_unlinked(i_am_reckless=True)
                        sp.statistics()
                        sp.write_database()
                    except:
                        print ('\nDropping unlinked exchanges failed')
            else:
                raise Warning ('\nStopped writing to backend SQLite3 database')      
        return datasets, exchanges


    @property
    def available_databases(self):
        '''All databases that have been loaded into ``Brightway2``.'''
        return bw2.databases

















#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
BW2QSD: Bridging Brightway2 and QSD packages for LCA

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/BW2QSD/blob/main/LICENSE.txt
for license details.
'''

__all__ = ('remove_setups_pickle', 'export_df','format_name',)

import os
from warnings import warn

def remove_setups_pickle():
    '''
    Remove the "setups.pickle" file in the project directory,
    outdated file will cause an error the next time ``BW2QSD`` is loaded.
    
    .. note::
        
        Only run this function before you quitting the program.
    
    '''
    
    from bw2data import projects
    setup_path = os.path.join(projects.dir, 'setups.pickle')
    try:
        os.remove(setup_path)
        print(f'File "setups.pickle" successfully removed from directory "{projects.dir}".')
    except FileNotFoundError:
        warn(f'"setups.pickle" not found in directory "{projects.dir}", no file removed.',
             stacklevel=2)


def export_df(df, path, show=False):
    if show:
        print(df)

    if path:
        if path.endswith('.tsv'):
            df.to_csv(path, sep='\t')
        elif path.endswith('.csv'):
            df.to_csv(path)
        elif (path.endswith('.xlsx') or path.endswith('.xls')):
            df.to_excel(path)
        else:
            extension = path.split('.')[-1]
            raise ValueError('Only "tsv", "csv", "xlsx", or "xls" files are supported, ' \
                             f'not {extension}.')

        file_path, file_name = os.path.split(path)
        print(f'File "{file_name}" has been exported to "{file_path}".')
    

def format_name(name):
    name = name.replace(', ', ' ')
    name = name.replace('-', ' ')
    return ''.join([i[0].capitalize() + i[1:] for i in name.split(' ')])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
BW2QSD: Bridging Brightway2 and QSD packages for LCA

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/BW2QSD/blob/main/LICENSE.txt
for license details.
'''

import pkg_resources, sys, os
try:
    __version__ = pkg_resources.get_distribution('bw2qsd').version
except pkg_resources.DistributionNotFound:
    __version__ = None

from ._exceptions import *

try:
    import brightway2 as bw2
except Exception as e:
    name, msg = str(sys.exc_info()[0]), str(sys.exc_info()[1])
    if 'ImportError' in name and "cannot import name 'databases'" in msg:
        raise BW2Error('This error is due to an outdated pickle file, ' \
                        'to get more instructions, update your ``bw2data`` package ' \
                        'according to https://github.com/brightway-lca/brightway2-data/commit/9c52e76c84bfa7d3d9719da152c0616d4039a3c3.')
    elif 'PickleError' in name and 'setups.pickle' in msg:
        print('\n')
        print('===================================================================')
        print('You can fix the following error by removing the setups.pickle file.')
        print('===================================================================')
        print('\n')
        print(e, '\n')
    
        print('Would you like to automatically remove the file? ' \
              'If yes, the file will be permanently removed ' \
              '(i.e., cannot be recovered from Trash or Recycle Bin).')
        if input('[y]/[n]: ') in ('y', 'Y', 'yes', 'Yes', 'YES'):
             path = msg.split("'")[-2]
             if not 'setups.pickle' in path:
                 print('\nError in parsing directory, for cautious purpose, ' \
                       f'please mannual remove the file in {msg} (always backup).\n')
                 raise e
             else:
                 os.remove(path)
                 print('\nFile successfully removed. Please rerun the code.\n')
                 sys.exit()
    
        else:
            raise e
    else:
        raise e

if 'biosphere3' not in bw2.databases:
    bw2.bw2setup()


from .utils import *
from ._db_downloader import *
from ._cf_getter import *

from . import (
    _exceptions,
    utils,
    _db_downloader,
    _cf_getter,    
    )

__all__ = (
    'remove_setup_pickle',
    *_db_downloader.__all__,
    *_cf_getter.__all__,
    )








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 26 14:10:52 2021

@author: Yalin Li
"""




# %%

# from bw2qsd import DataDownloader


# downloader = DataDownloader()
# downloader.download_ecoinvent()
# downloader.download_forwast()


# %%

from bw2qsd import CFgetter

ei = CFgetter('ei')
ei.load_database('ecoinvent_apos371')

ei.load_indicators(add=True, method='TRACI')

ei.load_activities('building', True)


# ei.remove('indicators', (('TRACI', 'environmental impact', 'acidification'),))

path = ''
# path = '/Users/yalinli_cabbi/OneDrive/Coding/BW2QSD/bw2qsd/sample_output.tsv'
df = ei.get_CF(path=path)



# %%

# Only run this at the very end
from bw2qsd import remove_setups_pickle
remove_setups_pickle()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
BW2QSD: Bridging Brightway2 and QSD packages for LCA

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/BW2QSD/blob/main/LICENSE.txt
for license details.
'''

'''
TODO:
    Improve search ability, update docs

'''


import sys, os
import pandas as pd
import brightway2 as bw2
from collections.abc import Iterable
from bw2data.backends.peewee import Activity
from .utils import export_df, format_name

isinstance = isinstance

__all__ = ('CFgetter',)


    
def _filter_inds(inds, cats, include):
    for n, cat in enumerate(cats):
        if cat:
            if isinstance(cat, str):
                cat = (cat, )
            elif not isinstance(cat, Iterable):
                raise TypeError(f'"{cat}" can only be str or iterable, ' \
                                f'not {type(cat).__name__}.')
            if include is True:
                for string in cat:
                    string_lower = string.lower()
                    inds = [ind for ind in inds if string_lower in ind[n].lower()]
            else:                
                for string in cat:
                    string_lower = string.lower()
                    inds = [ind for ind in inds if not string_lower in ind[n].lower()]
                
    return inds


class CFgetter:
    '''
    To get environmental impact characterization factors from databases through
    ``Brightway2``, users can choose different life cycle inventory (LCI) databases
    and life cycle impact assessment (LCIA).
    
    Tip
    ---
    Basic steps:
        
        [1] Load database (you can only load one at a time).
        
        [2] Load indicators (you can load multiple).
        
        [3] Load activities (you can load multiple).


    .. note::
        
        You need to first download the database using :class:`DataDownloader`.


    See Also
    --------
    `Brightway2 <2.docs.brightway.dev>`_
    
    '''

    __slots__ = ('name', '_database', '_indicators', '_activities', '_CFs')

    def __init__(self, name):
        self.name = name
        self._database = None
        self._indicators = set()
        self._activities = {}
        self._CFs = None

    def __repr__(self):
        return f'<CFgetter: {self.name}>'

    def load_database(self, database):
        '''
        Load designated database.
        
        Parameters
        ----------
        database: str
            Name of the database.        
        '''        
        db_lower = database.lower()
        
        if db_lower not in bw2.databases:
            raise ValueError(f'Database "{database}" not available. ' \
                             'Please first download the data using `DataDownloader`.')
        else:
            self._database = db = bw2.Database(database)

        print(f'Database {db} with {len(db)} inventories has been loaded.')


    def load_indicators(self, add=False, method='', method_exclude='',
                        category='', category_exclude='',
                        indicator='', indicator_exclude=''):
        '''
        Select and/or load designated impact indicators.
        
        Parameters
        ----------
        add : bool
            Whether to include the indicators in impact assessment for characterization factors.
            If False, a dict of the indicators that satisfy the search criteria
            will be returned.
        method: str or Iterable
            Impact assessment method of the indicator (e.g., TRACI).
        method_exclude: str or Iterable
            Strings to be excluded from the method field.
        category: str or Iterable
            Category of the indicator (e.g., environmental impact).
        category_exclude: str or Iterable
            Strings to be excluded from the category field.
        indicator: str or Iterable
            Name of the indicator (e.g., global warming).
        indicator_exclude: str or Iterable
            Strings to be excluded from the indicator field.

        Tip
        ---
        Leave all fields as blank (i.e., '') if want all of the options.
        E.g., load_indicators() will return
        all indicators available in the package (800+ in total).
        
        '''       
        indicators = bw2.methods.list
        indicators = _filter_inds(indicators, (method, category, indicator), True)
        indicators = _filter_inds(indicators, 
                                  (method_exclude, category_exclude, indicator_exclude),
                                  False)

        if add:
            self._indicators = self._indicators.union(set(indicators))
            msg = 'indicator' if len(indicators) > 1 else 'indicators'
            print(f'\n{len(indicators)} {msg} loaded/updated for {self.name}.')

        else:
            return indicators

    
    def load_activities(self, string: str, add: bool, limit=20, show=False,
                        **kwargs):
        '''
        Select and/or load activities of interest.

        Parameters
        ----------
        string : str
            Search string.
        add : bool
            Whether to include the activities in impact assessment for characterization factors.
            If False, a dict of the activities that satisfy the search criteria
            will be returned.
        limit : int
            Maximum number of search results to return.
        show : bool
            Whether to print the detailed information associated with the activities.
        kwargs :
            Other keyword arguments that will be passed to ```bw2data``.
        
        See Also
        --------
        :func:`search` in `bw2data SQLiteBackend <https://2.docs.brightway.dev/technical/bw2data.html#default-backend-databases-stored-in-a-sqlite-database>`_

        '''
        activities = self.database.search(string, limit=limit, **kwargs)
        act_dct = {act.as_dict()['name']: act for act in activities}

        for act in activities:
            if show:
                self.show_activity(act)
 
        if add:
            self._activities.update(act_dct)
            msg = 'activities' if len(act_dct) > 1 else 'activity'
            print(f'{len(act_dct)} {msg} loaded/updated for {self.name}.\n')

        else:
            return act_dct

    def show_activity(self, activity, **kwargs):
        '''
        Show detailed description about an activity.
        
        Parameters
        ----------
        activity : str or :class:`Activity`
            Name of the activity as str, or the :class:`Activity` in ``bw2data``.
        kwargs :
            Other keyword arguments that will be passed to ```bw2data``.

        .. note::
            
            When given as a str, the function will first try to search within
            the loaded activities, if there is no match,
            then it will search the entire database.

        '''
        if isinstance(activity, str):
            try:
                activities = [self.activities[activity]]
            except KeyError:
                try:
                    activities = [self.activities[activity]]
                except KeyError:
                    activities = self.database.search(activity, **kwargs)
                    if not activities:
                        raise ValueError(f'No search results for "{activity}".')
        
        else:
            if not isinstance(activity, Activity):
                raise TypeError(f'The input activity type "{type(activity).__name__}" is not valid.')
            else:
                activities = [activity]

        dfs = []
        for act in activities:
            act_dct = act.as_dict()
            name = act_dct['name']
            df = pd.DataFrame(
                data={name: [v for v in act_dct.values()]},
                index=[k for k in act_dct.keys()])
            print(f'{df.fillna("N/A")}\n')
            
            dfs.append(df)
        
        if len(dfs) == 1:
            return df        
        else:
            return dfs
        

    def remove(self, kind, keys):
        '''
        Remove loaded indicators or activities.
        
        Parameters
        ----------
        kind : str
            Can be "indicator" or "activity".
        keys : iterable    
            Key(s) of the item(s) to be removed.
        '''
        
        kind_lower = kind.lower()
        num = 0
        if kind_lower in ('indicator', 'indicators'):
            for k in keys:
                self._indicators.remove(k)
                num += 1
            msg = 'indicators' if num > 1 else 'indicator'

        elif kind_lower in ('activity', 'activities'):
            for k in keys:
                self._activities.pop(k)
                num += 1
            msg = 'activities' if num > 1 else 'activity'
        
        else:
            raise ValueError('kind can only be "indicator" or "activity", ' \
                             f'not "{kind}".')
    
        print(f'Successfully removed {num} {msg} from {self.name}.')
    
    def export_indicators(self, indicators=(), aliases={}, descriptions={},
                          show=False, path=''):
        '''
        Show information about the loaded impact indicators in a :class`pandas.DataFrame`,
        the :class`pandas.DataFrame` will be exported to the path if provided.
        
        Parameters
        ----------
        indicators : iterable
            Keys of the indicators in the ``indicators`` property.
            Will be defaulted to all loaded indicators if not provided.
        aliases : dict
            Keys should be the keys of the indicators in the ``indicators`` property,
            values should be the aliases of the indicators.
        descriptions : dict
            Keys should be the keys of the indicators in the ``indicators`` property,
            values should be the descriptions of the indicators.
        show : bool
            Whether to print the generated :class:`pandas.DataFrame` in the console.
        path : str
            If provided, the :class:`pandas.DataFrame` will be saved to the given file path.
        
        Returns
        -------
        df: :class:`pandas.DataFrame`
            Characterization factors.
            
        Tip
        ---
        The exported file can be directly import into QSD packages to create
        :class:`ImpactIndicator` items.
        
        
        '''
        if not self.indicators:
            raise ValueError('No loaded indicators.')
        if not set(indicators).issubset(self.indicators):
            raise ValueError('Provided indicator(s) not all loaded.')
        
        inds = indicators if indicators else self.indicators

        df = pd.DataFrame({
            'alias': [aliases[ind] if ind in aliases.keys() else '' for ind in inds],
            'unit': [bw2.methods.get(i)['unit'] for i in inds],
            'method': [ind[0] for ind in inds],
            'category': [ind[1] for ind in inds],
            'description': [descriptions[ind] if ind in descriptions.keys() else '' for ind in inds]
            },
            index=pd.Index(data=[format_name(ind[2]) for ind in inds], name='indicator'))
        
        export_df(df, path, show)

        return df
        
    

    def get_CF(self, indicators=(), activities=(), show=False, path=''):
        '''
        Get impact characterization factors.
        
        Parameters
        ----------
        indicators : iterable
            Keys of the indicators in the ``indicators`` property.
            Will be defaulted to all loaded indicators if not provided.
        activities : iterable
            Keys of the activities in the ``activities`` property.        
            Will be defaulted to all loaded indicators if not provided.
        show : bool
            Whether to print all characterization factors in the console.
        path : str
            If provided, the :class:`pandas.DataFrame` will be saved to the given file path.
        
        Returns
        -------
        df: :class:`pandas.DataFrame`
            Characterization factors.
        
        Tip
        ---
        [1] Use ``show_activity`` to see the functional unit of the activity.
        The quantity will be 1.
        
        [2] If you run into an "FileNotFoundError", most likely there are some
        indicators that do not acutally have corresponding impact assessment methods,
        try to load indicators one at a time to look for the culprit.
        
        '''
        
        if not self.indicators:
            raise ValueError('No loaded indicators.')
        elif not self.activities:
            raise ValueError('No loaded activities.')
        
        if not set(indicators).issubset(self.indicators):
            raise ValueError('Provided indicator(s) not all loaded.')
        
        inds = indicators if indicators else self.indicators
        ind_units = [bw2.methods.get(i)['unit'] for i in inds] + ['-']
        acts = [self.activities[k] for k in activities] if activities \
            else [v for v in self.activities.values()]
        
        bw2.calculation_setups['multiLCA'] = {'inv': [{a:1} for a in acts],
                                              'ia': inds}
        # This is to prevent bw2 from printing in the console
        stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')
        mlca = bw2.MultiLCA('multiLCA')
        sys.stdout = stdout

        pd_indices = [a['name'] for a in acts]
        pd_cols = pd.MultiIndex.from_tuples(inds, names=('method', 'category', 'indicator'))        
        cf_df = pd.DataFrame(data=mlca.results, index=pd_indices, columns=pd_cols)
        cf_df[('-', '-', 'activity functional unit')] = [a['unit'] for a in acts]

        unit_df = pd.DataFrame(data={k: v for k, v in zip(cf_df.columns, ind_units)},
                               index=['indicator unit'])

        df = pd.concat((unit_df, cf_df))
        df.sort_index(axis=1, inplace=True)
        
        if show:
            print(df)

        export_df(df, path, show)
        self._CFs = df

        return df

    def copy(self, name, omit=()):
        '''
        Return a copy of the :class:`CFgetter`.
        
        .. note:
            
            ``CFs`` is not copied.
        
        
        Parameters
        ----------
        name : str
            Name of the copy.
        omit: str or Iterable
            Property(ies) not to be copied.
        
        '''
        new = self.__class__.__new__(self.__class__)
        new.name = name
        new._CFs = None
        if isinstance(omit, str):
            omit = (omit,)
        omit = (*(f'_{o}' for o in omit), 'name', '_CFs')
        
        for s in self.__slots__:
            if not s in omit:
                if s == '_database':
                    new._database = self._database
                else:
                    setattr(new, s, getattr(self, s).copy())
        
        return new

    __copy__ = copy

    @property
    def available_databases(self):
        '''All databases that have been loaded into ``Brightway2``.'''
        return bw2.databases

    @property
    def database(self):
        '''Loaded database.'''
        return self._database
    @database.setter
    def database(self, i):
        raise AttributeError('Use ``load_database`` to load/switch database.')

    @property
    def indicators(self):
        '''[list] Loaded impact indicators.'''
        return list(self._indicators)
    @indicators.setter
    def indicators(self, i):
        raise AttributeError('Use ``load_indicators``/``remove`` to add/remove indicators.')

    @property
    def activities(self):
        '''[dict] Loaded activities.'''
        return self._activities
    @activities.setter
    def activities(self, i):
        raise AttributeError('Use ``load_activities``/``remove`` to add/remove activities.')

    @property
    def CFs(self):
        '''Results from :func:`get_CF`.'''
        if not self._CFs is None:
            return self._CFs
        else:
            try: return self.get_CF()
            except: return




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
BW2QSD: Bridging Brightway2 and QSD packages for LCA

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/BW2QSD/blob/main/LICENSE.txt
for license details.
'''

__all__ = ('BW2Error', )

class BW2Error(Exception):
    '''Errors related to Brightway2 packages.'''


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
BW2QSD: Bridging Brightway2 and QSD packages for LCA

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License. Please refer to
https://github.com/QSD-Group/BW2QSD/blob/main/LICENSE.txt
for license details.
'''

from setuptools import setup

setup(
    name='bw2qsd',
    packages=['bw2qsd'],
    version='0.0.8',
    license='UIUC',
    author='Yalin Li',
    author_email='zoe.yalin.li@gmail.com',
    description='Bridging Brightway2 and QSD packages for LCA',
    long_description=open('README.rst').read(),
    url="https://github.com/QSD-Group/BW2QSD",
    install_requires=['eidl', 'xlrd<=1.2.0'],
    package_data=
        {'bw2qsd': [
                    'sample_output.tsv',
                    'Tutorial.ipynb',
#                    'data/sanunit_data/*',
                    ]},
    platforms=['Windows', 'Mac', 'Linux'],
    classifiers=['License :: OSI Approved :: University of Illinois/NCSA Open Source License',
                 'Environment :: Console',
                 'Topic :: Education',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Chemistry',
                 'Topic :: Scientific/Engineering :: Mathematics',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Manufacturing',
                 'Intended Audience :: Science/Research',
                 'Natural Language :: English',
                 'Operating System :: MacOS',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX',
                 'Operating System :: POSIX :: BSD',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Unix',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 ],
    keywords=['Brightway2', 'quantitative sustainable design', 'life cycle assessment', 'database importer'],
)